#!/usr/bin/env python3

# Copyright 2016-2019 RnD Center "ELVEES", JSC

from argparse import Action
from argparse import ArgumentParser
import logging
import pprint
import sys
import threading
import time

import serial
import serial.rs485

KiB = 1024
MARKER = b'start\n'


class writeThread(threading.Thread):
    def __init__(self, tty, send_message):
        threading.Thread.__init__(self)
        self.send_message = send_message
        self.tty = tty
        self.write_time = None
        self.with_marker = False

    def run(self):
        start_time = time.time()
        if self.with_marker:
            self.tty.write(MARKER)
        self.tty.write(self.send_message)
        self.tty.flush()
        self.write_time = time.time() - start_time


class readThread(threading.Thread):
    def __init__(self, tty, message_to_check):
        threading.Thread.__init__(self)
        self.tty = tty
        self.message_to_check = message_to_check
        self.check_ok = False
        self.read_time = None
        self.receive_message = b''
        self.with_marker = False

    def run(self):
        message_wait_timeout = self.tty.timeout
        start_message_wait_time = time.time()
        while True:
            if (time.time() - start_message_wait_time) > message_wait_timeout:
                raise Exception('Message wait timeout')
            if self.tty.in_waiting:
                if self.with_marker:
                    self.tty.read(len(MARKER))
                start_time = time.time()
                self.receive_message = self.tty.read(len(self.message_to_check))
                break
        self.read_time = time.time() - start_time
        if self.receive_message == self.message_to_check:
            self.check_ok = True
        else:
            logging.error("receive message != send message")


def serial_stress_test(message,
                       port="/dev/ttyS1",
                       baudrate=1152000,
                       stopbits=1,
                       marker=False,
                       rs485_half_duplex=False,
                       rtscts=False):
    """Perform serial stress test and print statistics"""

    totalbytes = len(message)
    totalbits = len(message) * (8 + 1 + stopbits)  # 8 data bits + 1 start bit + stopbits
    estimate_read_time = totalbits / baudrate

    try:
        if rs485_half_duplex:
            tty = serial.rs485.RS485(port=port,
                                     baudrate=baudrate,
                                     stopbits=stopbits,
                                     timeout=estimate_read_time*1.5)
            tty.rs485_mode = serial.rs485.RS485Settings(rts_level_for_tx=True,
                                                        rts_level_for_rx=False)
        else:
            tty = serial.Serial(port=port,
                                baudrate=baudrate,
                                stopbits=stopbits,
                                timeout=estimate_read_time*1.5,
                                rtscts=rtscts)
    except Exception:
        logging.exception("Can't open device '%s'" % port)
        return None

    assert baudrate == tty.baudrate

    write_thread = writeThread(tty, message)
    write_thread.with_marker = marker
    read_thread = readThread(tty, message)
    read_thread.with_marker = marker

    start_time = time.time()
    write_thread.start()
    read_thread.start()

    write_thread.join()
    read_thread.join()
    elapsed_time = time.time() - start_time

    write_time = write_thread.write_time
    read_time = read_thread.read_time

    if not read_thread.check_ok:
        return None

    estimated_write_baudrate = totalbits / write_time
    estimated_read_baudrate = totalbits / read_time

    print("UART port configuration:")
    pprint.pprint(tty.get_settings(), indent=8)

    print("duplex mode : {}-duplex".format('half' if rs485_half_duplex else 'full'))
    print("total transferred data bytes : %d KiB" % (totalbytes/KiB))
    print("elapsed time : %f sec" % elapsed_time)

    print("estimated write baudrate (with stopbits): %d b/sec" % estimated_write_baudrate)
    print("estimated read baudrate (with stopbits): %d b/sec" % estimated_read_baudrate)
    print("write performance efficiency: %f %%" %
          ((estimated_write_baudrate/tty.baudrate) * 100))
    print("read performance efficiency: %f %%" %
          ((estimated_read_baudrate/tty.baudrate) * 100))

    return estimated_write_baudrate, estimated_read_baudrate


stopbits_map = {
    '1': serial.STOPBITS_ONE,
    '1.5': serial.STOPBITS_ONE_POINT_FIVE,
    '2': serial.STOPBITS_TWO}


class StopBitsAction(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, stopbits_map[values])


def main():
    description = "Client script for serial port testing"
    parser = ArgumentParser(description=description)

    parser.add_argument("-p", dest="port", default="/dev/ttyS1",
                        help="terminal where serial port is connected"
                             " (default: %(default)s)")

    parser.add_argument("-b", dest="baudrate", type=int,
                        default=115200,
                        help="baudrate (default: %(default)s)")

    parser.add_argument("-s", dest="stopbits",
                        choices=stopbits_map.keys(),
                        action=StopBitsAction,
                        default=serial.STOPBITS_ONE,
                        help="stop bits (default: %(default)s)")

    parser.add_argument("-t", dest="transfer_size", type=int,
                        default=1,
                        help="transfer size in KiB"
                             " (default: %(default)s)")

    parser.add_argument("--tap", action="store_true", help="show result in TAP")

    parser.add_argument("--min-efficiency", type=int, default=None,
                        help="minimum speed efficiency in percent")

    parser.add_argument("--rs485-half-duplex", action="store_true",
                        help="use RS485 port in half duplex mode")

    parser.add_argument("--rtscts", action="store_true",
                        help="use RTS and CTS pins for flow control")

    parser.add_argument("--marker", action="store_true",
                        help="use marker before data to prevent noise capture by server")

    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG)

    totalbytes = args.transfer_size * KiB
    send_message = b'1' * (totalbytes - 1) + b'\n'

    if args.tap:
        print("1..3")

    stress_test_result = serial_stress_test(message=send_message,
                                            port=args.port,
                                            baudrate=args.baudrate,
                                            stopbits=args.stopbits,
                                            marker=args.marker,
                                            rs485_half_duplex=args.rs485_half_duplex,
                                            rtscts=args.rtscts)

    write_efficiency = ((stress_test_result[0] if stress_test_result is not None else 0) /
                        args.baudrate) * 100
    read_efficiency = ((stress_test_result[1] if stress_test_result is not None else 0) /
                       args.baudrate) * 100
    is_write_efficiency_good = (args.min_efficiency is None) or \
        (write_efficiency >= args.min_efficiency)
    is_read_efficiency_good = (args.min_efficiency is None) or \
        (read_efficiency >= args.min_efficiency)
    if not is_write_efficiency_good and not args.tap:
        logging.error("Write efficiency is too small (%g%% but expected >= %d%%)" % (
            write_efficiency, args.min_efficiency))
    if not is_read_efficiency_good and not args.tap:
        logging.error("Read efficiency is too small (%g%% but expected >= %d%%)" % (
            read_efficiency, args.min_efficiency))

    if args.tap:
        OKNOTOK = {False: "not ok", True: "ok"}
        print("%s 1 transmit/receive data" % OKNOTOK[bool(stress_test_result)])
        if args.min_efficiency is not None:
            print("#write efficiency = %g%% (expected >= %d%%)" % (write_efficiency,
                                                                   args.min_efficiency))
            print("%s 2 write efficiency" % OKNOTOK[is_write_efficiency_good])
            print("#read efficiency = %g%% (expected >= %d%%)" % (read_efficiency,
                                                                  args.min_efficiency))
            print("%s 3 read efficiency" % OKNOTOK[is_read_efficiency_good])
        else:
            print("ok 2 write efficiency # SKIP checking efficiency not requested")
            print("ok 3 read efficiency # SKIP checking efficiency not requested")

    if stress_test_result and is_write_efficiency_good and is_read_efficiency_good:
        return 0

    return 1

if __name__ == "__main__":
    sys.exit(main())
