#!/usr/bin/env python3

# Copyright 2016-2019 RnD Center "ELVEES", JSC

from argparse import Action
from argparse import ArgumentParser
import collections
import logging

import serial
import serial.rs485

KiB = 1024
MARKER = b'start\n'


def wait_marker(tty):
    queue = collections.deque(maxlen=len(MARKER))
    while b''.join(queue) != MARKER:
        char = tty.read(1)
        queue.append(char)


def echo_server_start(transfer_size,
                      port="/dev/ttyS1",
                      baudrate=1152000,
                      stopbits=1,
                      timeout=None,
                      rs485_half_duplex=False,
                      marker=False):
    """Start endless receive -> translate loop """
    try:
        if rs485_half_duplex:
            tty = serial.rs485.RS485(port=port,
                                     baudrate=baudrate,
                                     stopbits=stopbits,
                                     timeout=timeout)
            tty.rs485_mode = serial.rs485.RS485Settings(rts_level_for_tx=True,
                                                        rts_level_for_rx=False,
                                                        delay_before_tx=0.1)
            # Default state of RTS signal is undefined. We must trigger write() function to
            # set correctly RTS state. After write() calling RS485 will be set to receive.
            tty.write(b'')
        else:
            tty = serial.Serial(port=port,
                                baudrate=baudrate,
                                stopbits=stopbits,
                                timeout=timeout)
    except Exception:
        logging.exception("Can't open device '%s'" % port)
        return None

    assert baudrate == tty.baudrate

    print("Echo server has been started!\n")

    while True:
        if marker:
            wait_marker(tty)

        message = tty.read(transfer_size)

        if marker:
            message = MARKER + message
        tty.write(message)

    return 0


stopbits_map = {
    '1': serial.STOPBITS_ONE,
    '1.5': serial.STOPBITS_ONE_POINT_FIVE,
    '2': serial.STOPBITS_TWO}


class StopBitsAction(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, stopbits_map[values])


def main():
    description = (
        "Echo server for serial port testing"
        "Server works in two modes:"
        "* full duplex"
        "* half duplex"
        "Full duplex is an infinite loop receiving and transmitting transfer-size bytes from given"
        "port."
        "Half duplex is an infinite loop receiving and transmitting transfer-size bytes from given"
        "port with delay 1 sec between receiving and transmitting")
    parser = ArgumentParser(description=description)

    parser.add_argument("-p", dest="port", default="/dev/ttyS1",
                        help="terminal where serial port is connected"
                             " (default: %(default)s)")

    parser.add_argument("-b", dest="baudrate", type=int,
                        default=115200,
                        help="baudrate (default: %(default)s)")

    parser.add_argument("-s", dest="stopbits",
                        choices=stopbits_map.keys(),
                        action=StopBitsAction,
                        default=serial.STOPBITS_ONE,
                        help="stop bits (default: %(default)s)")

    parser.add_argument("-t", dest="transfer_size",
                        type=float,
                        default=0.0625,  # Equal to 64 Bytes
                        help="transfer size in KiB (default: %(default)s)")

    parser.add_argument("--rs485-half-duplex", action="store_true",
                        help="use RS485 port in half duplex mode")

    parser.add_argument("--marker", action="store_true",
                        help="wait marker before data to prevent noise capture")

    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG)

    echo_server_start(port=args.port,
                      baudrate=args.baudrate,
                      stopbits=args.stopbits,
                      transfer_size=int(args.transfer_size * KiB),
                      rs485_half_duplex=args.rs485_half_duplex,
                      marker=args.marker)


if __name__ == "__main__":
    main()
